import {Pipe} from "../support/Pipe";

type CollectionItem<T> = T
type MaybeCollectionItem<T> = CollectionItem<T> | undefined
type KeyFunction<T, T2> = (item: CollectionItem<T>, index: number) => CollectionItem<T2>
type KeyReducerFunction<T, T2> = (current: any, item: CollectionItem<T>, index: number) => T2
type CollectionFunction<T, T2> = (items: Collection<T>) => T2
type KeyOperator<T, T2> = string | KeyFunction<T, T2>
type AssociatedCollectionItem<T2, T> = { key: T2, item: CollectionItem<T> }
type CollectionComparable<T> = CollectionItem<T>[] | Collection<T>
type DeterminesEquality<T> = (item: CollectionItem<T>, other: any) => boolean
type CollectionIndex = number
type MaybeCollectionIndex = CollectionIndex | undefined
type ComparisonFunction<T> = (item: CollectionItem<T>, other_item: CollectionItem<T>) => number

import { WhereOperator, applyWhere, whereMatch } from "./where";

const collect = <T>(items: CollectionItem<T>[]): Collection<T> => Collection.collect(items)
export {
    collect,
    Collection,

    // Types
    CollectionItem,
    MaybeCollectionItem,
    KeyFunction,
    KeyReducerFunction,
    CollectionFunction,
    KeyOperator,
    AssociatedCollectionItem,
    CollectionComparable,
    DeterminesEquality,
    CollectionIndex,
    MaybeCollectionIndex,
    ComparisonFunction,
}

/**
 * A helper class for working with arrays of items in a more robust fashion.
 * Provides helpers for accessing sub-keys, filtering, piping, and aggregate functions.
 */
class Collection<T> {
    private _items: CollectionItem<T>[] = []

    /**
     * Create a new collection from an array of items.
     * @param items
     */
    public static collect<T>(items: CollectionItem<T>[]): Collection<T> {
        return new Collection(items)
    }

    /**
     * Create a collection of "undefined" elements of a given size.
     * @param size
     */
    public static size(size: number): Collection<undefined> {
        const arr = Array(size).fill(undefined)
        return new Collection<undefined>(arr)
    }

    /**
     * Fill a new collection of the given size with the given item.
     * @param size
     * @param item
     */
    public static fill<T>(size: number, item: T): Collection<T> {
        const arr = Array(size).fill(item)
        return new Collection<T>(arr)
    }

    constructor(
        /**
         * The items to base the collection on.
         */
        items?: CollectionItem<T>[]
    ) {
        if ( items )
            this._items = items
    }

    private _all<T2>(key: KeyOperator<T, T2>): CollectionItem<T2>[] {
        let items: CollectionItem<T2>[] = []
        if ( typeof key === 'function' ) {
            items = this._items.map(key)
        } else {
            items = this._items.map((item: CollectionItem<T>) => (<any>item)[key])
        }
        return items
    }

    private _all_numbers<T2>(key: KeyOperator<T, T2>): number[] {
        return this._all(key).map(value => Number(value))
    }

    private _all_associate<T2>(key: KeyOperator<T, T2>): AssociatedCollectionItem<T2, T>[] {
        const assoc_items: AssociatedCollectionItem<T2, T>[] = []
        let items = [...this._items]
        if ( typeof key === 'function' ) {
            items.map((item, index) => {
                const key_item = key(item, index)
                assoc_items.push({ key: key_item, item })
            })
        } else {
            items.map((item, index) => {
                assoc_items.push({key: (<any>item)[key], item})
            })
        }
        return assoc_items
    }

    /**
     * Cast the collection to an array.
     */
    all(): CollectionItem<T>[] {
        return [...this._items]
    }

    /**
     * Get the average value of the items or one of their keys.
     * @param key
     */
    average<T2>(key?: KeyOperator<T, T2>): number {
        let items
        if ( key ) items = this._all_numbers(key)
        else items = this._items.map(x => Number(x))
        if ( items.length === 0 ) return 0

        let sum = items.reduce((prev, curr) => prev + curr)
        return sum / items.length
    }

    /**
     * Get the median value of the items or one of their keys.
     * @param key
     */
    median<T2>(key?: KeyOperator<T, T2>): number {
        let items
        if ( key ) items = this._all_numbers(key).sort((a, b) => a - b)
        else items = this._items.map(x => Number(x)).sort((a, b) => a - b)

        const middle = Math.floor((items.length - 1) / 2)
        if ( items.length % 2 ) return items[middle]
        else return (items[middle] + items[middle + 1]) / 2
    }

    /**
     * Get the mode of the items or one of their keys.
     * @param key
     */
    mode<T2>(key?: KeyOperator<T, T2>): number {
        let items
        if ( key ) items = this._all_numbers(key).sort((a, b) => a - b)
        else items = this._items.map(x => Number(x)).sort((a, b) => a - b)

        let counts: any = {}
        for ( const item of items ) {
            counts[item] = (counts[item] ?? -1) + 1
        }

        return Number(Object.keys(counts).reduce((a, b) => counts[a] > counts[b] ? a : b)[0])
    }

    /**
     * Collapse a (potentially nested) collection of items down to a single dimension.
     */
    collapse(): Collection<any> {
        const new_items: CollectionItem<T>[] = []
        const items = [...this._items]
        const get_layer = (current: CollectionItem<T>|CollectionItem<T>[]) => {
            if ( typeof (<any>current)[Symbol.iterator] === 'function' ) {
                // @ts-ignore
                // TODO fix this
                for (const item of current) {
                    if (Array.isArray(item)) {
                        get_layer(item)
                    } else {
                        new_items.push(item)
                    }
                }
            }
        }

        get_layer(items)
        return new Collection(new_items)
    }

    /**
     * Returns true if the given key matches the given condition for any item in the collection.
     *
     * @example
     * ```typescript
     * const userExists = users.contains('username', '=', 'jdoe')
     * ```
     *
     * @param key
     * @param operator
     * @param operand
     */
    contains<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): boolean {
        const associate = this._all_associate(key)
        const matches = applyWhere(associate, operator, operand)
        return matches.length > 0
    }

    // TODO crossJoin

    /**
     * Create a copy of this collection.
     * Does NOT deep copy the underlying items.
     */
    clone(): Collection<T> {
        return new Collection<T>(this._items)
    }

    /**
     * Return a collection of items that ARE in this collection, but NOT in the `items` collection.
     * @param items
     */
    diff<T2>(items: CollectionComparable<T|T2>): Collection<T> {
        const exclude = items instanceof Collection ? items.all() : items
        const matches = []
        for ( const item of [...this._items] ) {
            if ( !exclude.includes(item) ) matches.push(item)
        }
        return new Collection(matches)
    }

    /**
     * Return a collection of items that ARE in this collection, but NOT In the `items` collection
     * using a helper function to determine whether two items are equal.
     *
     * @example
     * ```typescript
     * potentialUsers.diffUsing(storedUsers, (u1, u2) => u1.username.toLowerCase() === u2.username.toLowerCase())
     * ```
     *
     * @param items
     * @param compare
     */
    diffUsing<T2>(items: CollectionComparable<T|T2>, compare: DeterminesEquality<T>): Collection<T> {
        const exclude = items instanceof Collection ? items.all() : items
        const matches = []
        for ( const item of [...this._items] ) {
            if ( !exclude.some(exc => compare(item, exc)) )
                matches.push(item)
        }
        return new Collection(matches)
    }

    /**
     * Returns true if the given function returns truthy for any item in the collection.
     * Stops executing if a single truth case is found.
     * @param func
     */
    some(func: (item: T) => boolean): boolean {
        return this._items.some(func)
    }

    /**
     * Execute the function for every item in the collection.
     * @param func
     */
    each<T2>(func: KeyFunction<T, T2>): void {
        this._items.map(func)
    }

    /**
     * Create a new collection by mapping the items in this collection using the given function.
     * @param func
     */
    map<T2>(func: KeyFunction<T, T2>): Collection<T2> {
        const new_items: CollectionItem<T2>[] = []
        this.each(((item, index) => {
            new_items.push(func(item, index))
        }))
        return new Collection<T2>(new_items)
    }

    /**
     * Convert this collection to an object keyed by the given field.
     *
     * @example
     * ```typescript
     * const users = collect([{uid: 1, name: 'John'}, {uid: 2, name: 'Jane'}])
     * users.keyBy('name')  // => {John: {uid: 1, name: 'John'}, Jane: {uid: 2, name: 'Jane'}}
     * ```
     *
     * @param key
     */
    keyBy(key: KeyOperator<T, string>): {[key: string]: T} {
        const obj: {[key: string]: T} = {}

        this._all_associate(key).forEach(assoc => {
            obj[assoc.key] = assoc.item
        })

        return obj
    }

    /**
     * Convert this collection to an object keyed by the given field, whose values are
     * the output of the `value` operator.
     *
     * @example
     * ```typescript
     * const users = collect([{uid: 1, name: 'John'}, {uid: 2, name: 'Jane'}])
     * users.keyMap('name', 'uid')  // => {John: 1, Jane: 2}
     * ```
     *
     * @param key
     * @param value
     */
    keyMap<T2>(key: KeyOperator<T, string>, value: KeyOperator<T, T2>): {[key: string]: T2} {
        const obj: {[key: string]: T2} = {}

        let i = -1;
        this._all_associate(key).forEach(assoc => {
            i += 1;

            if ( typeof value === 'string' ) {
                // @ts-ignore
                obj[assoc.key] = assoc.item[value]
            } else {
                obj[assoc.key] = value(assoc.item, i)
            }
        })

        return obj
    }

    /**
     * Returns true if the given function returns true for every item in the collection.
     * @param func
     */
    every<T2>(func: KeyFunction<T, T2>): boolean {
        return this._items.every(func)
    }

    everyWhere<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): boolean {
        const items = this._all_associate(key)
        return items.every(item => whereMatch(item, operator, operand))
    }

    /**
     * Return a new collection filtered by the given function.
     * @param func
     */
    filter<T2>(func: KeyFunction<T, T2>): Collection<T> {
        return new Collection(this._items.filter(func))
    }

    /**
     * Returns the index of the record for which the given function returns true, if such an index exists.
     * @param func
     */
    find<T2>(func: KeyFunction<T, T2>): number | undefined {
        let found_index: number | undefined = undefined
        this._items.some((item, index) => {
            if ( func(item, index) ) {
                found_index = index
                return true
            }
        })
        return found_index
    }

    /**
     * When `bool` is truthy, execute the callback, passing in the collection.
     * Can be used for functional-style chained calls.
     * @param bool
     * @param then
     */
    when<T2>(bool: boolean, then: CollectionFunction<T, T2>): Collection<T> {
        if ( bool ) then(this)
        return this
    }

    /**
     * When `bool` is falsy, execute the callback, passing in the collection.
     * Can be used for functional-style chained calls.
     * @param bool
     * @param then
     */
    unless<T2>(bool: boolean, then: CollectionFunction<T, T2>): Collection<T> {
        if ( !bool ) then(this)
        return this
    }

    /**
     * Filter the collection by the given where-condition.
     *
     * @example
     * ```typescript
     * const users = collect([
     *  {uid: 1, name: 'John'},
     *  {uid: 2, name: 'Jane'},
     *  {uid: 3, name: 'James'},
     * ])
     *
     * users.where('uid', '<', 3)  // => Collection[{uid: 1, name: 'John'}, {uid: 2, name: 'Jane'}]
     * ```
     *
     * @param key
     * @param operator
     * @param operand
     */
    where<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): Collection<T> {
        const items = this._all_associate(key)
        return new Collection(applyWhere(items, operator, operand))
    }

    /**
     * Filter the collection by the inverse of the given where-condition.
     * @param key
     * @param operator
     * @param operand
     */
    whereNot<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): Collection<T> {
        return this.diff(this.where(key, operator, operand))
    }

    /**
     * Filter the collection for all records where the given key is in a set of items.
     * @param key
     * @param items
     */
    whereIn<T2>(key: KeyOperator<T, T2>, items: CollectionComparable<T2>): Collection<T> {
        const allowed = items instanceof Collection ? items.all() : items
        const matches = []
        for ( const { key: search, item } of this._all_associate(key) ) {
            if ( allowed.includes(search) ) matches.push(item)
        }
        return new Collection(matches)
    }

    /**
     * Filter the collection for all records where the given key is NOT in a set of items.
     * @param key
     * @param items
     */
    whereNotIn<T2>(key: KeyOperator<T, T2>, items: CollectionComparable<T2>): Collection<T> {
        return this.diff(this.whereIn(key, items))
    }

    /**
     * Return the first item in the collection, if it exists.
     */
    first(): MaybeCollectionItem<T> {
        if ( this.length > 0 ) return this._items[0]
    }

    /**
     * Return the first item in the collection that matches the given where-condition.
     * @param key
     * @param operator
     * @param operand
     */
    firstWhere<T2>(key: KeyOperator<T, T2>, operator: WhereOperator = '=', operand: any = true): MaybeCollectionItem<T> {
        const items = this.where(key, operator, operand).all()
        if ( items.length > 0 ) return items[0]
    }

    /**
     * Return the first item in the collection that does NOT match the given where-condition.
     * @param key
     * @param operator
     * @param operand
     */
    firstWhereNot<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): MaybeCollectionItem<T> {
        const items = this.whereNot(key, operator, operand).all()
        if ( items.length > 0 ) return items[0]
    }

    /**
     * Get the number of items in the collection.
     */
    get length() {
        return this._items.length
    }

    /**
     * Get the number of items in the collection.
     */
    count() {
        return this._items.length
    }

    // TODO flatten - depth

    /**
     * Get the item at the given index in the collection. If none exists,
     * return the fallback.
     * @param index
     * @param fallback
     */
    get(index: number, fallback?: any) {
        if ( this.length > index ) return this._items[index]
        else return fallback
    }

    /**
     * Return the item at the given index in the collection, if it exists.
     * @param index
     */
    at(index: number): MaybeCollectionItem<T> {
        return this.get(index)
    }

    /**
     * Return an object mapping key values to arrays of records with that key.
     *
     * @example
     * ```typescript
     * const users = collect([
     *  {uid: 1, name: 'John', type: 'admin'},
     *  {uid: 2, name: 'Jane', type: 'user'},
     *  {uid: 3, name: 'James', type: 'user'},
     * ])
     *
     * users.groupBy('type')  // => {admin: [{uid: 1, ...}], user: [{uid: 2, ...}, {uid: 3, ...}]}
     * ```
     *
     * @param key
     */
    groupBy<T2>(key: KeyOperator<T, T2>): any {
        const items = this._all_associate(key)
        const groups: any = {}
        for ( const item of items ) {
            const key = String(item.key)
            if ( !groups[key] ) groups[key] = []
            groups[key].push(item.item)
        }
        return groups
    }

    /**
     * Return an object mapping the given key to the record with that key.
     * See `keyBy()`.
     * @param key
     */
    associate<T2>(key: KeyOperator<T, T2>): any {
        const items = this._all_associate(key)
        const values: any = {}
        for ( const item of items ) {
            values[String(item.key)] = item.item
        }
        return values
    }

    /**
     * Join the items in the collection to a string delimited by the given delimiter.
     * @param delimiter
     */
    join(delimiter: string): string {
        return this._items.join(delimiter)
    }

    /**
     * Join the items in the collection to a string delimited by the given delimiter.
     * @param delimiter
     */
    implode(delimiter: string): string {
        return this.join(delimiter)
    }

    /**
     * Return a collection of the items that exist in both this collection and this collection,
     * (optionally) using the given key to compare.
     * @param items
     * @param key
     */
    intersect(items: CollectionComparable<T>, key?: KeyOperator<T, T>): Collection<T> {
        const compare = items instanceof Collection ? items.all() : items
        const intersect = []
        let all_items
        if ( key ) all_items = this._all_associate(key)
        else all_items = this._items.map(item => {
            return { key: item, item }
        })

        for ( const item of all_items ) {
            if ( compare.includes(item.key) )
                intersect.push(item.item)
        }
        return new Collection(intersect)
    }

    /**
     * True if the collection has no items.
     */
    isEmpty(): boolean {
        return this.length < 1
    }

    /**
     * True if the collection has at least one item.
     */
    isNotEmpty(): boolean {
        return this.length > 0
    }

    /**
     * Return the last item in the collection.
     */
    last(): MaybeCollectionItem<T> {
        if ( this.length > 0 ) return this._items.reverse()[0]
    }

    /**
     * Return the last item in the collection that matches the given where-condition.
     * @param key
     * @param operator
     * @param operand
     */
    lastWhere<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): MaybeCollectionItem<T> {
        const items = this.where(key, operator, operand).all()
        if ( items.length > 0 ) return items.reverse()[0]
    }

    /**
     * Return the last item in the collection that does NOT match the given where-condition.
     * @param key
     * @param operator
     * @param operand
     */
    lastWhereNot<T2>(key: KeyOperator<T, T2>, operator: WhereOperator, operand?: any): MaybeCollectionItem<T> {
        const items = this.whereNot(key, operator, operand).all()
        if ( items.length > 0 ) return items.reverse()[0]
    }

    /**
     * Map the collection to a collection of the values of the key.
     *
     * @example
     * ```typescript
     * const users = collect([
     *  {uid: 1, name: 'John', type: 'admin'},
     *  {uid: 2, name: 'Jane', type: 'user'},
     *  {uid: 3, name: 'James', type: 'user'},
     * ])
     *
     * users.pluck('name')  // => Collection['John', 'Jane', 'James']
     * ```
     *
     * @param key
     */
    pluck<T2>(key: KeyOperator<T, T2>): Collection<T2> {
        return new Collection<T2>(this._all(key))
    }

    /**
     * Return the max value of the given key.
     * @param key
     */
    max<T2>(key: KeyOperator<T, T2>): number {
        const values = this._all_numbers(key)
        return Math.max(...values)
    }

    /**
     * Return the item with the max value of the given key.
     * @param key
     */
    whereMax<T2>(key: KeyOperator<T, T2>): Collection<T> {
        return this.where(key, '=', this.max(key))
    }

    /**
     * Return the min value of the given key.
     * @param key
     */
    min<T2>(key: KeyOperator<T, T2>): number {
        const values = this._all_numbers(key)
        return Math.min(...values)
    }

    /**
     * Return the item with the min value of the given key.
     * @param key
     */
    whereMin<T2>(key: KeyOperator<T, T2>): Collection<T> {
        return this.where(key, '=', this.min(key))
    }

    /**
     * Get a new collection containing both the items in this collection, and the `items` collection.
     * @param items
     */
    merge<T2>(items: CollectionComparable<T2>): Collection<T|T2> {
        const merge = items instanceof Collection ? items.all() : items
        return new Collection([...this._items, ...merge])
    }

    /**
     * Return every nth item in the collection.
     *
     * @example
     * ```
     * const items = collect(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i'])
     *
     * items.nth(3)  // => Collection['a', 'd', 'g']
     * ```
     *
     * @param n
     */
    nth(n: number): Collection<T> {
        const matches: CollectionItem<T>[] = []
        let current = 1
        this._items.forEach((item, index) => {
            if ( current === 1 ) matches.push(item)
            current += 1
            if ( current > n ) current = 1
        })
        return new Collection(matches)
    }

    /**
     * Return the items that should exist on the given page, assuming there are `perPage` many items on a single page.
     * @param page
     * @param perPage
     */
    forPage(page: number, perPage: number) {
        const start = page * perPage - perPage
        const end = page * perPage
        return new Collection(this._items.slice(start, end))
    }

    /**
     * Return a new Pipe of this collection.
     */
    pipe(): Pipe<Collection<T>> {
        return Pipe.wrap(this)
    }

    /**
     * Remove the last item from this collection.
     */
    pop(): MaybeCollectionItem<T> {
        if ( this.length > 0 ) {
            return this._items.pop()
        }
    }

    /**
     * Add the given item to the beginning of this collection.
     * @param item
     */
    prepend(item: CollectionItem<T>): Collection<T> {
        this._items = [item, ...this._items]
        return this
    }

    /**
     * Add the given item to the end of this collection.
     * @param item
     */
    push(item: CollectionItem<T>): Collection<T> {
        this._items.push(item)
        return this
    }

    /**
     * Push the given items to the end of this collection.
     * Unlike `merge()`, this mutates the current collection's items.
     * @param items
     */
    concat(items: CollectionComparable<T>): Collection<T> {
        const concats = items instanceof Collection ? items.all() : items
        for ( const item of concats ) {
            this._items.push(item)
        }
        return this
    }

    /**
     * Insert the given item into this collection at the specified index.
     * @param index
     * @param item
     */
    put(index: number, item: CollectionItem<T>): Collection<T> {
        const new_items = []
        let inserted = false
        this._items.forEach((existing, existing_index) => {
            if ( existing_index === index ) {
                new_items.push(item)
                inserted = true
            }

            new_items.push(existing)
        })

        if ( !inserted ) new_items.push(item)
        return new Collection(new_items)
    }

    /**
     * Return `n` many randomly-selected items from this collection.
     * @param n
     */
    random(n: number): Collection<T> {
        const random_items: CollectionItem<T>[] = []
        const all = this._items
        if ( n > this.length ) n = this.length
        while ( random_items.length < n ) {
            const item = all[Math.floor(Math.random() * all.length)]
            if ( !random_items.includes(item) ) random_items.push(item)
        }
        return new Collection(random_items)
    }

    /**
     * Reduce this collection to a single value using the given reducer function.
     *
     * @example
     * ```typescript
     * const items = collect([1, 3, 5, 7])
     *
     * items.reduce((sum, item) => sum + item, 0)  // => 16
     * ```
     *
     * @param reducer
     * @param initial_value
     */
    reduce<T2>(reducer: KeyReducerFunction<T, T2>, initial_value?: T2): T2 | undefined {
        let current_value = initial_value
        this._items.forEach((item, index) => {
            current_value = reducer(current_value, item, index)
        })
        return current_value
    }

    /**
     * Return a new collection of items that fail the given truth-test function.
     * @param truth_test
     */
    reject<T2>(truth_test: KeyFunction<T, T2>): Collection<T> {
        const rejected = this._items.filter((item, index) => {
            return !truth_test(item, index)
        })
        return new Collection(rejected)
    }

    /**
     * Return a new collection whose items are in the reverse order of the current one.
     */
    reverse(): Collection<T> {
        return new Collection([...this._items.reverse()])
    }

    /**
     * Try to find the given item in the collection. If it exists, return the index.
     * @param item
     */
    search(item: CollectionItem<T>): MaybeCollectionIndex {
        let found_index
        this._items.some((possible_item, index) => {
            if ( possible_item === item ) {
                found_index = index
                return true
            }
        })
        return found_index
    }

    /**
     * Remove and return the first item in the collection, if it exists.
     */
    shift(): MaybeCollectionItem<T> {
        if ( this.length > 0 ) {
            return this._items.shift()
        }
    }

    /**
     * Shuffle the items in this collection to a random order.
     */
    shuffle(): Collection<T> {
        const items = [...this._items]
        for ( let i = items.length - 1; i > 0; i-- ) {
            const j = Math.floor(Math.random() * (i + 1))
            ;[items[i], items[j]] = [items[j], items[i]]
        }
        return new Collection(items)
    }

    /**
     * Return a sub-set of this collection between the given index ranges.
     * @param start
     * @param end
     */
    slice(start: number, end: number) {
        return new Collection(this._items.slice(start, end))
    }

    // TODO split
    // TODO chunk

    /**
     * Sort the collection (optionally) using the given comparison function.
     * @param compare_func
     */
    sort(compare_func?: ComparisonFunction<T>): Collection<T> {
        const items = this._items
        if ( compare_func ) items.sort(compare_func)
        else items.sort()
        return new Collection(items)
    }

    /**
     * Sort the collection (optionally) using the given key operator.
     * @param key
     */
    sortBy<T2>(key?: KeyOperator<T, T2>): Collection<T> {
        let items: any[]
        if ( key ) items = this._all_associate(key)
        else items = this._items.map(item => {
            return { key: item, item }
        })

        items.sort((a: any, b: any) => {
            if ( a.key > b.key ) return 1
            else if ( a.key < b.key ) return -1
            else return 0
        })
        return new Collection(items.map((item: AssociatedCollectionItem<T2, T>) => item.item))
    }

    /**
     * Identical to `sort()`, but in reverse order.
     * @param compare_func
     */
    sortDesc(compare_func?: ComparisonFunction<T>): Collection<T> {
        return this.sort(compare_func).reverse()
    }

    /**
     * Identical to `sortBy()`, but in reverse order.
     * @param key
     */
    sortByDesc<T2>(key?: KeyOperator<T, T2>): Collection<T> {
        return this.sortBy(key).reverse()
    }

    /**
     * Remove `deleteCount` many items from the collection, starting at the `start` index.
     * @param start
     * @param deleteCount
     */
    splice(start: CollectionIndex, deleteCount?: number): Collection<T> {
        return new Collection([...this._items].splice(start, deleteCount))
    }

    /**
     * Return the sum of the items in the collection, optionally by key.
     *
     * @example
     * ```typescript
     * const items = collect([{k1: 1}, {k1: 3}, {k1: 5}, {k1: 7}])
     *
     * items.sum('k1')  // => 16
     * ```
     *
     * @param key
     */
    sum<T2>(key?: KeyOperator<T, T2>): number {
        let items
        if ( key ) items = this._all_numbers(key)
        else items = this._items.map(x => Number(x))
        return items.reduce((prev, curr) => prev + curr)
    }

    /**
     * Return a collection of the first `limit` many items in this collection.
     * If `limit` is negative, returns the last `limit` many items.
     * @param limit
     */
    take(limit: number): Collection<T> {
        if ( limit === 0 ) return new Collection()
        else if ( limit > 0 ) {
            return new Collection(this._items.slice(0, limit))
        } else {
            return new Collection(this._items.reverse().slice(0, -1 * limit).reverse())
        }
    }

    /**
     * Apply the given function to this collection then return the collection.
     * This is intended to help with chaining.
     *
     * @example
     * ```typescript
     * collection.tap(coll => {
     *         coll.push(item)
     *     })
     *     .where('someKey', '>', 4)
     *     // ... &c.
     * ```
     *
     * @param func
     */
    tap<T2>(func: CollectionFunction<T, T2>): Collection<T> {
        func(this)
        return this
    }

    /**
     * Return all distinct items in this collection. If a key is specified, returns
     * all unique values of that key.
     *
     * @example
     * ```typescript
     * const users = collect([
     *  {uid: 1, name: 'John', type: 'admin'},
     *  {uid: 2, name: 'Jane', type: 'user'},
     *  {uid: 3, name: 'James', type: 'user'},
     * ])
     *
     * users.unique('type')  // => Collection['admin', 'user']
     * ```
     *
     * @param key
     */
    unique<T2>(key?: KeyOperator<T, T2>): Collection<T|T2> {
        const has: CollectionItem<T|T2>[] = []
        let items
        if ( key ) items = this._all<T2>(key)
        else items = [...this._items]
        for ( const item of items ) {
            if ( !has.includes(item) ) has.push(item)
        }
        return new Collection(has)
    }

    /**
     * Returns true if the given item is in this collection.
     * @param item
     */
    includes(item: CollectionItem<T>): boolean {
        return this._items.includes(item)
    }

    /**
     * Add on to the end of this collection as many `value` items as necessary until the collection is `length` long.
     * @param length
     * @param value
     */
    pad(length: number, value: CollectionItem<T>): Collection<T> {
        const items = [...this._items]
        while ( items.length < length ) items.push(value)
        return new Collection(items)
    }

    /**
     * Cast the collection to an array.
     */
    toArray(): any[] {
        const returns: any = []
        for ( const item of this._items ) {
            if ( item instanceof Collection ) returns.push(item.toArray())
            else returns.push(item)
        }
        return returns
    }

    /**
     * Cast the collection to a JSON string, optionally specifying the replacer and indentation.
     * @param replacer
     * @param space
     */
    toJSON(replacer = undefined, space = 4): string {
        return JSON.stringify(this.toArray(), replacer, space)
    }

    // TODO getIterator
    // TODO getCachingIterator

    [Symbol.iterator]() {
        const items = this._items
        return {
            current_index: 0,
            next() {
                if ( items.length < 1 || this.current_index + 1 > items.length ) {
                    return { done: true }
                }

                const item = items[this.current_index]
                this.current_index += 1
                return { done: false, value: item }
            },
        }
    }

    /**
     * Like `map()`, but the callback can be async.
     *
     * @example
     * A trivial example, but demonstrative:
     *
     * ```typescript
     * const collection = collect([1, 2, 3])
     *
     * collection.map(async item => item + 1)  // => Collection[Promise<1>, Promise<2>, Promise<3>]
     *
     * collection.promiseMap(async item => item + 1)  // => Promise<Collection[2, 3, 4]>
     * ```
     *
     * @param func
     */
    async promiseMap<T2>(func: KeyFunction<T, T2 | Promise<T2>>): Promise<Collection<T2>> {
        return new Collection<T2>(await Promise.all(
            this.map(func).toArray()
        ))
    }
}
